"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import json
import os
import subprocess
import sys
from abc import ABC
from urllib.parse import unquote, urlparse

from gofigr import CodeLanguage
from gofigr.context import RevisionContext

PATH_WARNING = "To fix this warning, you can manually specify the notebook name & path in the call to configure(). " \
               "Please see https://gofigr.io/docs/gofigr-python/latest/customization.html#notebook-name-path " \
               "for details."


class Annotator(ABC):
    """\
    Annotates figure revisions with pertinent information, such as cell code, variable values, etc.

    """
    def __init__(self, extension):
        self.extension = extension

    def annotate(self, revision):
        """
        Annotates the figure revision.

        :param revision: FigureRevision
        :return: annotated FigureRevision

        """
        return revision


class CellIdAnnotator(Annotator):
    """Annotates revisions with the ID of the Jupyter cell"""
    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        try:
            cell_id = self.extension.cell.cell_id
        except AttributeError:
            cell_id = None

        revision.metadata['cell_id'] = cell_id

        return revision


class CellCodeAnnotator(Annotator):
    """"Annotates revisions with cell contents"""
    def annotate(self, revision):
        if self.extension.cell is not None:
            code = self.extension.cell.raw_cell
        else:
            code = "N/A"

        revision.data.append(revision.client.CodeData(name="Jupyter Cell",
                                                      language=CodeLanguage.PYTHON,
                                                      contents=code))
        return revision


class PipFreezeAnnotator(Annotator):
    """Annotates revisions with the output of pip freeze"""
    def __init__(self, extension, cache=True):
        """\
        :param extension: the GoFigr Jupyter extension
        :param cache: if True, will only run pip freeze once and cache the output
        """
        super().__init__(extension)
        self.cache = cache
        self.cached_output = None

    def annotate(self, revision):
        if self.cache and self.cached_output:
            output = self.cached_output
        else:
            try:
                output = subprocess.check_output(["pip", "freeze"]).decode('ascii')
                self.cached_output = output
            except subprocess.CalledProcessError as e:
                output = e.output

        revision.data.append(revision.client.TextData(name="pip freeze", contents=output))
        return revision


class SystemAnnotator(Annotator):
    """Annotates revisions with the OS version"""
    def annotate(self, revision):
        try:
            output = subprocess.check_output(["uname", "-a"]).decode('ascii')
        except subprocess.CalledProcessError as e:
            output = e.output

        revision.data.append(revision.client.TextData(name="System Info", contents=output))
        return revision


NOTEBOOK_PATH = "notebook_path"
NOTEBOOK_NAME = "notebook_name"
NOTEBOOK_URL = "url"
NOTEBOOK_KERNEL = "kernel"
PYTHON_VERSION = "python_version"
BACKEND_NAME = "backend"


class NotebookMetadataAnnotator(Annotator):
    """"Annotates revisions with notebook metadata, including filename & path, as well as the full URL"""

    def parse_metadata(self):
        """Infers the notebook path & name from metadata passed through the WebSocket (if available)"""
        meta = self.extension.notebook_metadata
        if meta is None:
            raise RuntimeError("No Notebook metadata available")
        if 'url' not in meta:
            raise RuntimeError("No URL found in Notebook metadata")

        notebook_name = unquote(urlparse(meta['url']).path.rsplit('/', 1)[-1])
        notebook_dir = self.extension.shell.starting_dir
        full_path = os.path.join(notebook_dir, notebook_name)

        if not os.path.exists(full_path):
            print(f"The inferred path for the notebook does not exist: {full_path}. {PATH_WARNING}", file=sys.stderr)

        return {NOTEBOOK_PATH: full_path,
                NOTEBOOK_NAME: notebook_name,
                NOTEBOOK_URL: meta.get('url')}

    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        try:
            if NOTEBOOK_NAME not in revision.metadata or NOTEBOOK_PATH not in revision.metadata:
                revision.metadata.update(self.parse_metadata())
        except Exception:  # pylint: disable=broad-exception-caught
            print(f"GoFigr could not automatically obtain the name of the currently"
                  f" running notebook. {PATH_WARNING}",
                  file=sys.stderr)

            revision.metadata[NOTEBOOK_NAME] = "N/A"
            revision.metadata[NOTEBOOK_PATH] = "N/A"

        return revision


class NotebookNameAnnotator(NotebookMetadataAnnotator):
    """(Deprecated) Annotates revisions with notebook name & path"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("NotebookNameAnnotator is deprecated. Please use NotebookMetadataAnnotator", file=sys.stderr)


class EnvironmentAnnotator(Annotator):
    """Annotates revisions with the python version & the kernel info"""
    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        revision.metadata[NOTEBOOK_KERNEL] = sys.executable
        revision.metadata[PYTHON_VERSION] = sys.version

        return revision


class BackendAnnotator(Annotator):
    """Annotates revisions with the python version & the kernel info"""
    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        context = RevisionContext.get(revision)
        revision.metadata[BACKEND_NAME] = context.backend.get_backend_name() if context and context.backend else "N/A"

        return revision


class HistoryAnnotator(Annotator):
    """Annotates revisions with IPython execution history"""
    def annotate(self, revision):
        context = RevisionContext.get(revision)

        if not hasattr(context.extension.shell, 'history_manager'):
            return revision

        hist = context.extension.shell.history_manager
        if hist is None:
            return revision

        revision.data.append(revision.client.CodeData(name="IPython history",
                                                      language="python",
                                                      format="jupyter-history/json",
                                                      contents=json.dumps(hist.input_hist_raw)))
        return revision



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import time
from datetime import datetime, timedelta

import numpy as np

from gofigr import UnauthorizedError
from gofigr.models import WorkspaceMembership

from tests.test_client import MultiUserTestCase


def has_invite(workspace, invite):
    for inv in workspace.get_invitations():
        if inv.api_id == invite.api_id:
            return True
    return False


class TestInvitations(MultiUserTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n_revisions = 0

    def test_creation_and_deletion(self):
        gf = self.gf1
        invite = gf.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                        workspace=gf.primary_workspace,
                                        membership_type=WorkspaceMembership.CREATOR)
        invite.create()
        self.assertIsNotNone(invite.token)

        all_invites = gf.primary_workspace.get_invitations()
        for inv in all_invites:
            self.assertIsNone(inv.token)  # should only be provided at creation

        self.assertIsNone(gf.WorkspaceInvitation(api_id=invite.api_id).fetch().token)
        self.assertIsNotNone(gf.WorkspaceInvitation(api_id=invite.api_id).fetch().email)

        self.assertTrue(has_invite(gf.primary_workspace, invite), msg="Created invite not found")

        workspace2 = gf.Workspace(name="Second workspace").create()
        invite2 = gf.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                         workspace=workspace2,
                                         membership_type=WorkspaceMembership.ADMIN)
        invite2.create()
        self.assertIsNotNone(invite2.token)
        self.assertNotEqual(invite.token, invite2.token)

        # Invites should be exclusive to workspaces
        self.assertTrue(has_invite(gf.primary_workspace, invite))
        self.assertFalse(has_invite(workspace2, invite))

        self.assertFalse(has_invite(gf.primary_workspace, invite2))
        self.assertTrue(has_invite(workspace2, invite2))

        # Delete both invites
        invite.delete()
        invite2.delete()
        self.assertFalse(has_invite(gf.primary_workspace, invite))
        self.assertFalse(has_invite(workspace2, invite))

        self.assertFalse(has_invite(gf.primary_workspace, invite2))
        self.assertFalse(has_invite(workspace2, invite2))

    def test_self_acceptance(self):
        invite1 = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                               workspace=self.gf1.primary_workspace,
                                               membership_type=WorkspaceMembership.CREATOR)
        invite1.create()

        # This should fail, because we're already a member
        self.assertRaises(RuntimeError, lambda: invite1.accept())

        # Invitations can only be tried once. This one should no longer exist.
        self.assertRaises(RuntimeError, lambda: invite1.fetch())

    def test_acceptance(self):
        invite = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                              workspace=self.gf1.primary_workspace,
                                              membership_type=WorkspaceMembership.CREATOR)
        token = invite.create().token

        # Verify that User 2 doesn't have access yet
        self.assertRaises(RuntimeError, lambda: self.gf2.Workspace(api_id=self.gf1.primary_workspace.api_id).fetch())

        # User 2 accepts the invite
        invite_u2 = self.gf2.WorkspaceInvitation(token=token)
        invite_u2.accept()

        # User 2 should now be listed as a member
        member, = [m for m in self.gf1.primary_workspace.get_members() if m.username == self.gf2.username]
        self.assertEqual(member.username, self.gf2.username)
        self.assertEqual(member.membership_type, WorkspaceMembership.CREATOR)

        # User 2 should have access to the workspace
        w = self.gf2.Workspace(api_id=self.gf1.primary_workspace.api_id).fetch()
        self.assertEqual(w.name, self.gf1.primary_workspace.name)

        # The invite should no longer exist
        self.assertRaises(RuntimeError, lambda: invite_u2.fetch())

    def test_expiration(self):
        invite = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                              workspace=self.gf1.primary_workspace,
                                              expiry=datetime.now() + timedelta(seconds=1),
                                              membership_type=WorkspaceMembership.CREATOR)
        token = invite.create().token

        # Wait for invitation to expire
        time.sleep(2)

        # User 2 tries to accept the invite. It should raise an exception.
        invite_u2 = self.gf2.WorkspaceInvitation(token=token)
        self.assertRaises(RuntimeError, lambda: invite_u2.accept())

    def test_malicious_invites(self):
        # User 1 invites themselves to User 2's workspace
        invite = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                              workspace=self.gf2.primary_workspace,
                                              expiry=datetime.now() + timedelta(seconds=1),
                                              membership_type=WorkspaceMembership.CREATOR)
        self.assertRaises(UnauthorizedError, lambda: invite.create())

        # User 2 tries to obtain and delete a valid invite created by User 1 without a token
        invite = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                              workspace=self.gf1.primary_workspace,
                                              expiry=datetime.now() + timedelta(seconds=1),
                                              membership_type=WorkspaceMembership.CREATOR)
        invite.create()

        invite2 = self.gf2.WorkspaceInvitation(api_id=invite.api_id)
        self.assertRaises(UnauthorizedError, lambda: invite2.fetch())
        self.assertRaises(UnauthorizedError, lambda: invite2.delete())

    def test_token_uniqueness(self):
        """This is a simple sanity check that tokens are random"""
        invite = self.gf1.WorkspaceInvitation(email="testuser@flagstaff.ai",
                                              workspace=self.gf1.primary_workspace,
                                              membership_type=WorkspaceMembership.CREATOR)
        token = invite.create().token

        # Verbatim token should work
        invite2 = self.gf1.WorkspaceInvitation(token=token).fetch()

        # Minor variations should not
        for _ in range(100):
            token_variation = list(token)
            rand_idx = np.random.choice(np.arange(len(token)))
            token_variation[rand_idx] = np.random.choice(list(set(token) - {token[rand_idx]}))
            self.assertRaises(RuntimeError,
                              lambda: self.gf1.WorkspaceInvitation(token="".join(token_variation)).fetch())



"""\
Copyright (c) 2022, Flagstaff Solutions, LLC
All rights reserved.

"""
# Many of the properties here are defined dynamically based on Field specifications, so no-member isn't meaningful
# pylint: disable=no-member

import abc
import io
from base64 import b64encode, b64decode
from collections import namedtuple
from http import HTTPStatus
from urllib.parse import urljoin
from uuid import uuid4

import PIL
import dateutil.parser

import pandas as pd

from gofigr.profile import MeasureExecution


class Field:
    """\
    Describes a dynamically created object field, i.e. figure name, revision etc.

    """
    def __init__(self, name, parent=None, derived=False):
        """\

        :param name: name of the field as it will appear in the parent object, i.e. object.<name>
        :param parent: instance of the object containing the field
        :param derived: whether the field is derived. Derived fields are not included in REST API calls.

        """
        self.name = name
        self.parent = parent
        self.derived = derived

    def to_representation(self, value):
        """Converts the value of this field to a JSON-serializable primitive"""
        return value

    def to_internal_value(self, gf, data):  # pylint: disable=unused-argument
        """\
        Parses the value of this field from JSON primitives.

        :param gf: GoFigr instance
        :param data: parsed JSON primitives
        :return: parsed value

        """
        return data

    def clone(self):
        """\
        Creates a clone of this field.

        :return: cloned field
        """
        return Field(self.name, parent=self.parent, derived=self.derived)


class JSONField(Field):
    """\
    Represents a field that stores JSON primitives.
    """
    def to_representation(self, value):
        return value

    def to_internal_value(self, gf, data):
        # This function is called on the result of response.json() which would have already parsed
        # nested fields.
        return data

    def clone(self):
        return JSONField(self.name, parent=self.parent, derived=self.derived)


class Timestamp(Field):
    """\
    Timestamp field
    """
    def to_representation(self, value):
        return str(value) if value is not None else None

    def to_internal_value(self, gf, data):
        return dateutil.parser.parse(data) if data is not None else None

    def clone(self):
        return Timestamp(self.name, parent=self.parent, derived=self.derived)


class Base64Field(Field):
    """\
    Timestamp field
    """
    def to_representation(self, value):
        return None if value is None else b64encode(value).decode('ascii')

    def to_internal_value(self, gf, data):
        return None if data is None else b64decode(data)

    def clone(self):
        return Base64Field(self.name, parent=self.parent, derived=self.derived)


class LinkedEntityCollection:
    """Represents a collection of linked entities, i.e. figures inside an analysis"""
    def __init__(self, entities, read_only=False, backlink_property=None, backlink=None):
        """\

        :param entities: list of entities
        :param read_only: if True, you won't be able to create new entities
        :param backlink_property: name of the property in each linked entity which will reference the parent object. \
        For example, if this collection stores figure revisions and backlink_property = "figure", you will \
        be able to refer to the parent figure as revision.figure.
        :param backlink: the parent object that backlink_property will point to
        """
        self._entities = list(entities)
        self.read_only = read_only
        self.backlink_property = backlink_property
        self.backlink = backlink

    def __iter__(self):
        return iter(self._entities)

    def __getitem__(self, item):
        return self._entities.__getitem__(item)

    def __repr__(self):
        return repr(self._entities)

    def __len__(self):
        return len(self._entities)

    def find(self, **kwargs):
        """\
        Returns the first object whose attributes match the query. E.g. find(name='hello', age=21) will return
        all objects where obj.name == "hello" and obj.age == 21.

        :param kwargs: query
        :return: first object that matches the query or None
        """
        for obj in self:
            if all(getattr(obj, name) == val for name, val in kwargs.items()):
                return obj

        return None

    def find_or_create(self, default_obj=None, **kwargs):
        """\
        Finds an object that matches query parameters. If the object doesn't exist, it will persist default_obj
        and return it instead.

        :param default_obj: object to create/persist if no matches are found
        :param kwargs: query parameters. See .find()
        :return: found or created object.

        """
        obj = self.find(**kwargs)
        if obj is not None:
            return obj
        elif default_obj is not None:
            self.create(default_obj)
            return default_obj
        else:
            raise RuntimeError(f"Could not find object: {kwargs}")

    def create(self, new_obj, **kwargs):
        """\
        Creates a new object and appends it to the collection.

        :param new_obj: object to create
        :return: created object

        """
        if self.read_only:
            raise RuntimeError("This collection is read only. Cannot create a new object.")

        if self.backlink_property is not None:
            setattr(new_obj, self.backlink_property, self.backlink)

        new_obj.create(**kwargs)
        self._entities.append(new_obj)
        return new_obj


class LinkedEntityField(Field):
    """\
    Represents a linked entity (or a collection of them), e.g. an Analysis inside a Workspace.
    """
    # pylint: disable=too-many-arguments
    def __init__(self, name, entity_type, many=False,
                 read_only=False, backlink_property=None, parent=None,
                 derived=False, sort_key=None):
        """\

        :param name: field name
        :param entity_type: type of the linked entity, e.g. Analysis, Figure, etc.
        :param many: if False, will create a single linked entity. If True, will resolve to a LinkedEntityCollection
        :param read_only: Makes the collection of linked entities read-only. Only relevant if many=True.
        :param backlink_property: name of property in the linked entity which will point back to the parent
        :param parent: parent object
        :param derived: True if derived (won't be transmitted through the API)
        :param sort_key: sort key (callable) for entities. None if no sort (default).

        """
        super().__init__(name, parent=parent, derived=derived)
        self.entity_type = entity_type
        self.many = many
        self.read_only = read_only
        self.backlink_property = backlink_property
        self.sort_key = sort_key

    def clone(self):
        return LinkedEntityField(self.name, self.entity_type,
                                 many=self.many, read_only=self.read_only,
                                 backlink_property=self.backlink_property,
                                 parent=self.parent, derived=self.derived,
                                 sort_key=self.sort_key)

    def to_representation(self, value):
        if value is None:
            return None
        elif self.many:
            sorted_value = value if self.sort_key is None else sorted(value, key=self.sort_key)
            return [x.api_id for x in sorted_value]
        else:
            return value.api_id

    def to_internal_value(self, gf, data):
        make_prefetched = lambda obj: self.entity_type(gf)(parse=True, **obj)
        make_not_prefetched = lambda api_id: self.entity_type(gf)(api_id=api_id)
        make_one = lambda obj: make_not_prefetched(obj) if isinstance(obj, str) else make_prefetched(obj)

        if data is None:
            return None
        elif self.many:
            sorted_vals = [make_one(api_id) for api_id in data]
            if self.sort_key is not None:
                sorted_vals = sorted(sorted_vals, key=self.sort_key)
            return LinkedEntityCollection(sorted_vals,
                                          read_only=self.read_only,
                                          backlink_property=self.backlink_property,
                                          backlink=self.parent)
        else:
            return make_one(data)

class DataField(LinkedEntityField):
    """Customizes a LinkedEntityField so that the entity data is fully embedded in the representation"""
    def clone(self):
        return DataField(self.name, self.entity_type,
                         many=self.many, read_only=self.read_only,
                         backlink_property=self.backlink_property,
                         parent=self.parent, derived=self.derived,
                         sort_key=self.sort_key)

    def _assert_not_shallow(self, obj):
        if obj.api_id is not None and obj.data is None:  # pylint: disable=protected-access
            raise ValueError("This is a shallow Data object without any data. It cannot be saved. "
                             "Please call fetch() first.")

        return True

    def to_representation(self, value):
        if value is None:
            return None
        elif self.many:
            sorted_value = value if self.sort_key is None else sorted(value, key=self.sort_key)
            return [x.to_json() for x in sorted_value if self._assert_not_shallow(x)]
        else:
            self._assert_not_shallow(value)
            return value.to_json()


class NestedEntityField(Field):
    """\
    Represents a nested entity. Nested entities are embedded fully in the parent object's JSON representation
    and cannot be manipulated on their own.
    """
    def __init__(self, name, entity_type, many=False, read_only=False, derived=False, parent=None):
        super().__init__(name, parent=parent, derived=derived)
        self.entity_type = entity_type
        self.many = many
        self.read_only = read_only

    def clone(self):
        return NestedEntityField(self.name, self.entity_type,
                                 many=self.many,
                                 read_only=self.read_only,
                                 derived=self.derived,
                                 parent=self.parent)

    def to_representation(self, value):
        if value is None:
            return None
        elif self.many:
            return [x.to_json() for x in value]
        else:
            return value.to_json()

    def to_internal_value(self, gf, data):
        make_one = lambda d: self.entity_type(gf).from_json(d)
        if self.many:
            coltype = tuple if self.read_only else list
            return coltype([make_one(d) for d in data])
        else:
            return make_one(data)


class NestedMixin(abc.ABC):
    """\
    Nested objects: these are not standalone (cannot be manipulated based on API ID), but are directly embedded
    inside other objects.
    """
    def to_json(self):
        """Converts this object to JSON"""
        raise NotImplementedError

    @classmethod
    def from_json(cls, data):
        """Parses this object from JSON"""
        raise NotImplementedError


class ModelMixin(abc.ABC):
    """Base class for GoFigr API entities: workspaces, analyses, figures, etc."""
    # pylint: disable=protected-access
    fields = ['api_id', "_shallow"]
    endpoint = None
    _gf = None  # GoFigr instance. Will be set dynamically.

    def __init__(self, api_id=None, parse=False, **kwargs):
        """

        :param api_id: API ID
        :param parse: if True, the fields' to_internal_value will be called on all properties. Otherwise, properties \
        will be stored verbatim. This is to support direct object creation from Python (i.e. parse = False), and \
        creation from JSON primitives (i.e. parse = True).
        :param kwargs:

        """
        self.api_id = api_id
        fields = [Field(x) if isinstance(x, str) else x.clone()
                  for x in self.fields]
        for fld in fields:
            fld.parent = self
        self.fields = {fld.name: fld for fld in fields}

        # Set all fields to None by default
        for name in self.fields:
            if not hasattr(self, name):
                setattr(self, name, None)

        self._update_properties(kwargs, parse=parse)

    def __eq__(self, other):
        repr1 = self.to_json()
        repr2 = other.to_json()

        if not set(repr1.keys()) == set(repr2.keys()):
            return False
        else:
            return all(repr1[k] == repr2[k] for k in repr1.keys()
                       if not isinstance(self.fields[k], Timestamp))

    def __hash__(self):
        raise RuntimeError("Model instances are not hashable")

    @classmethod
    def from_json(cls, data):
        """Parses an instance of this class from JSON data"""
        return cls(**data, parse=True)

    def to_json(self, include_derived=True, include_none=False):
        """\
        Serializes this model to JSON primitives.

        :param include_derived: if True, derived fields will be included in the representation
        :param include_none: if True, fields set to None will be included in the representation
        :return: this object (and all nested/referenced fields) serialized to JSON primitives.
        """
        data = {fld.name: fld.to_representation(getattr(self, fld.name, None))
                for fld in self.fields.values()
                if (not fld.derived or include_derived)}
        if not include_none:
            data = {k: v for k, v in data.items() if v is not None}

        return data

    def _update_properties(self, props, parse=True):
        for name, val in props.items():
            if name not in self.fields:
                continue

            field = self.fields[name]
            setattr(self, name, field.to_internal_value(self._gf, val) if parse else val)

        return self

    @property
    def client(self):
        """Returns the underlying GoFigr instance"""
        return self._gf

    @classmethod
    def list(cls):
        """\
        Lists all objects from the server.

        """
        res = cls._gf._get(cls.endpoint)
        return [cls(**obj, parse=True) for obj in res.json()]

    def save(self, create=False, patch=False, silent=False):
        """\
        Saves this object to server

        :param create: will create the object if it doesn't already exist. Otherwise, saving a non-existing object \
        will throw an exception.
        :param patch: if True, will submit a partial update where some required properties may be missing. \
        You will almost never use this: it's only useful if for some reason you can't/don't want to fetch the full \
        object before updating properties. However, the web app relies on this functionality so it's available.
        :param silent: if True, the server will not generate an activity for this update.
        :return: self
        """
        if self.api_id is None:
            if create:
                return self.create(update=False)
            else:
                raise RuntimeError("API ID is None. Did you forget to create() the object first?")

        if silent:
            params = "?silent=true"
        else:
            params = ""

        method = self._gf._patch if patch else self._gf._put
        with MeasureExecution("Save request"):
            response = method(urljoin(self.endpoint, self.api_id) + "/" + params,
                              json=self.to_json(include_derived=False),
                              expected_status=HTTPStatus.OK)

        with MeasureExecution("Parse response"):
            self._update_properties(response.json())
        return self

    def _check_api_id(self):
        if self.api_id is None:
            raise RuntimeError("API ID is None")

    def fetch(self):
        """\
        Updates all fields from the server. Note that any unsaved local changes will be overwritten.

        :return: self
        """
        self._check_api_id()
        obj = self._gf._get(urljoin(self.endpoint, self.api_id + "/")).json()
        return self._update_properties(obj)

    def create(self, update=False):
        """\
        Creates this object on the server.

        :param update: if True and the object already exists, its properties will be updated on the server. Otherwise
        trying to create an object which already exists will throw an exception.

        :return: self
        """
        if self.api_id is not None:
            if update:
                return self.save()
            else:
                raise RuntimeError("This entity already exists. Cannot create.")

        response = self._gf._post(self.endpoint, json=self.to_json(include_derived=False),
                                  expected_status=HTTPStatus.CREATED)
        self._update_properties(response.json())
        return self

    def delete(self, **kwargs):
        """\
        Deletes an object on the server. This cannot be undone.

        :param kwargs: specify delete=True to actually delete the object.
        :return:

        """
        self._check_api_id()

        if 'delete' not in kwargs or not kwargs['delete']:
            raise RuntimeError("Specify delete=True to delete this object. This cannot be undone.")

        return self._gf._delete(urljoin(self.endpoint, self.api_id + "/"))

    def __repr__(self):
        return str(self.to_json())


class LinkSharingStatus(NestedMixin):
    """Stores the status of link sharing for shareable objects."""
    def __init__(self, enabled):
        self.enabled = enabled

    @classmethod
    def from_json(cls, data):
        return LinkSharingStatus(enabled=data.get('enabled', False))

    def to_json(self):
        return {'enabled': self.enabled}

    def __repr__(self):
        return str(self.to_json())


class SharingUserData(NestedMixin):
    """Stores information about a user that an object has been shared with"""
    def __init__(self, username, sharing_enabled):
        if username is None:
            raise ValueError("Username cannot be None.")
        elif sharing_enabled is None:
            raise ValueError("sharing_enabled cannot be None")

        self.username = username
        self.sharing_enabled = sharing_enabled

    @classmethod
    def from_json(cls, data):
        return SharingUserData(username=data.get('username'),
                               sharing_enabled=data.get('sharing_enabled'))

    def to_json(self):
        return {'username': self.username,
                'sharing_enabled': self.sharing_enabled}

    def __repr__(self):
        return str(self.to_json())


class WorkspaceMember(NestedMixin):
    """Stores information about a member of a workspace"""
    def __init__(self, username, membership_type):
        if username is None:
            raise ValueError("Username cannot be None")

        self.username = username
        self.membership_type = membership_type

    @classmethod
    def from_json(cls, data):
        return WorkspaceMember(username=data.get('username'),
                               membership_type=data.get('membership_type'))

    def to_json(self):
        return {'username': self.username,
                'membership_type': self.membership_type}

    def __repr__(self):
        return str(self.to_json())


class LogItem(NestedMixin):
    # pylint: disable=too-many-instance-attributes, too-many-arguments

    """\
    Represents an activity item, such as a figure being created or modified.

    """
    def __init__(self, username, timestamp, action, target_id, target_type,
                 deleted_sentinel=None,
                 deleted=False,
                 thumbnail=None,
                 target_name=None,
                 analysis_id=None,
                 analysis_name=None,
                 api_id=None,
                 gf=None,
                 parent=None):
        """\

        :param username: user who performed the activity
        :param timestamp: time the activity was performed
        :param action: Type of action (a string): create, create_child, view, update, move, delete
        :param target_id: API ID of the target object
        :param target_type: type of the target object
        :param deleted_sentinel: if this is a delete action (target no longer exists), this field captures the name of
        the entity that was deleted
        :param deleted: True if target was deleted
        :param thumbnail: base-64 encoded thumbnail image of the target object
        :param target_name: not used
        :param analysis_id: ID of the parent analysis
        :param analysis_name: name of the parent analysis
        :param api_id: API ID for this activity
        :param gf: GoFigr client instance
        :param parent: parent object, e.g. Workspace or Analysis
        """
        self.username, self.timestamp, self.action = username, timestamp, action
        self.target_id, self.target_type = target_id, target_type
        self.deleted, self.deleted_sentinel = deleted, deleted_sentinel
        self.thumbnail = thumbnail
        self.target_name = target_name
        self.analysis_id = analysis_id
        self.analysis_name = analysis_name
        self.api_id = api_id
        self.gf = gf
        self.parent = parent

    def fetch(self):
        """Fetches information about this log item from the server. API ID and parent have to be set."""
        if self.api_id is None:
            raise ValueError("API ID is None")
        elif self.parent is None:
            raise ValueError("Parent is None")
        elif self.parent.api_id is None:
            raise ValueError("Parent API ID is None")

        # pylint: disable=protected-access
        obj = self.gf._get(urljoin(self.parent.endpoint + "/" + self.parent.api_id + "/log/", self.api_id + "/")).json()

        if 'timestamp' in obj.keys():
            obj['timestamp'] = dateutil.parser.parse(obj['timestamp'])

        for name, value in obj.items():
            setattr(self, name, value)

        return self

    @classmethod
    def from_json(cls, data, gf=None, parent=None):
        timestamp = data.get('timestamp')
        if timestamp:
            timestamp = dateutil.parser.parse(timestamp)

        return LogItem(username=data.get('username'),
                       timestamp=timestamp,
                       action=data.get('action'),
                       target_id=data.get('target_id'),
                       target_type=data.get('target_type'),
                       target_name=data.get('target_name'),
                       deleted_sentinel=data.get('deleted_sentinel'),
                       deleted=data.get('deleted'),
                       thumbnail=data.get('thumbnail'),
                       analysis_id=data.get('analysis_id'),
                       analysis_name=data.get('analysis_name'),
                       api_id=data.get('api_id'),
                       gf=gf,
                       parent=parent)

    def to_json(self):
        return {'username': self.username,
                'timestamp': self.timestamp,
                'action': self.action,
                'target_id': self.target_id,
                'target_type': self.target_type,
                'target_name': self.target_name,
                'deleted_sentinel': self.deleted_sentinel,
                'thumbnail': self.thumbnail,
                'deleted': self.deleted,
                'analysis_id': self.analysis_id,
                'analysis_name': self.analysis_name,
                'api_id': self.api_id}

    def __repr__(self):
        return str(self.to_json())


class ShareableModelMixin(ModelMixin):
    """\
    Mixins for things we can share: analyses, figures, revisions, etc.

    """
    # pylint: disable=protected-access

    def set_link_sharing(self, enabled):
        """\
        Enabled or disables link sharing.

        :param enabled: true to enable, false to disable.
        :return: confirmation of link sharing status (true or false)
        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/share/link/'),
                                  json=LinkSharingStatus(enabled=enabled).to_json(),
                                  expected_status=HTTPStatus.OK)
        return LinkSharingStatus(response.json()).enabled

    def get_link_sharing(self):
        """\
        Gets current status of link sharing

        :return: true if link sharing is enabled, false otherwise.
        """
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/share/link/'),
                                 expected_status=HTTPStatus.OK)
        return LinkSharingStatus.from_json(response.json()).enabled

    def share(self, username):
        """\
        Shares this object with a specific user.

        :param username: name of the user to share with.
        :return: SharingUserData object

        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/share/user/'),
                                  json=SharingUserData(username=username, sharing_enabled=True).to_json(),
                                  expected_status=HTTPStatus.OK)
        return SharingUserData.from_json(response.json())

    def unshare(self, username):
        """\
        Unshares this object from a user.

        :param username: username of a user from whom to remove access
        :return: SharingUserData

        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/share/user/'),
                                  json=SharingUserData(username=username, sharing_enabled=False).to_json(),
                                  expected_status=HTTPStatus.OK)
        return SharingUserData.from_json(response.json())

    def get_sharing_users(self):
        """\
        Gets a list of all users with whom this object has been shared.

        :return: list of SharingUserData objects.

        """
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/share/user/'),
                                 expected_status=HTTPStatus.OK)
        return [SharingUserData.from_json(datum) for datum in response.json()]


TIMESTAMP_FIELDS = ["created_by", Timestamp("created_on"),
                    "updated_by", Timestamp("updated_on")]

CHILD_TIMESTAMP_FIELDS = ["child_updated_by", Timestamp("child_updated_on")]


class WorkspaceType(abc.ABC):
    """Enum for workspace type"""
    SECONDARY = "secondary"
    PRIMARY = "primary"


class WorkspaceMembership(abc.ABC):
    """Enum for levels of workspace membership: owner, viewer, etc."""
    OWNER = "owner"
    ADMIN = "admin"
    CREATOR = "creator"
    VIEWER = "viewer"

    ALL_LEVELS = [OWNER, ADMIN, CREATOR, VIEWER]


Recents = namedtuple("Recents", ["analyses", "figures"])


class LogsMixin:
    """\
    Mixin for entities which support the /log/ endpoint.

    """
    def get_logs(self):
        """\
        Retrieves the activity log.

        :return: list of LogItem objects.
        """
        # pylint: disable=protected-access
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/log/'),
                                 expected_status=HTTPStatus.OK)
        return [LogItem.from_json(datum, gf=self._gf, parent=self) for datum in response.json()]


class ThumbnailMixin:
    """\
    Mixin for entities which support the /log/ endpoint.

    """
    def get_thumbnail(self, size):
        """\
        Retrieves the thumbnail image.
        :param size: size in pixels

        :return: PIL.Image if available or None
        """
        # pylint: disable=protected-access
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/thumbnail/{size}'),
                                 expected_status=HTTPStatus.OK)

        data = response.json()
        thumb_b64 = data.get('thumbnail')

        if not thumb_b64:
            return None

        img_data = b64decode(thumb_b64)
        bio = io.BytesIO(img_data)
        img = PIL.Image.open(bio)
        img.load()
        return img


class gf_Workspace(ModelMixin, LogsMixin):
    """Represents a workspace"""
    # pylint: disable=protected-access

    fields = ["api_id",
              "_shallow",
              "name",
              "description",
              "workspace_type",
              "size_bytes",
              LinkedEntityField("analyses", lambda gf: gf.Analysis, many=True, derived=True,
                                backlink_property='workspace')] + TIMESTAMP_FIELDS + CHILD_TIMESTAMP_FIELDS
    endpoint = "workspace/"

    def get_analysis(self, name, create=True, **kwargs):
        """\
        Finds an analysis by name.

        :param name: name of the analysis
        :param create: whether to create an analysis if one doesn't exist
        :param kwargs: if an Analysis needs to be created, parameters of the Analysis object (such as description)
        :return: Analysis instance.
        """
        return self.analyses.find_or_create(name=name,
                                            default_obj=self._gf.Analysis(name=name, **kwargs) if create else None)

    def get_invitations(self):
        """\
        Gets members of this workspace.

        :return: list of WorkspaceMember objects
        """
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/invitations/'),
                                 expected_status=HTTPStatus.OK)
        return [self._gf.WorkspaceInvitation(**datum, parse=True) for datum in response.json()]

    def get_members(self):
        """\
        Gets members of this workspace.

        :return: list of WorkspaceMember objects
        """
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/members/'),
                                 expected_status=HTTPStatus.OK)
        return [WorkspaceMember.from_json(datum) for datum in response.json()]

    def add_member(self, username, membership_type):
        """\
        Adds a member to this workspace.

        :param username: username of the person to add
        :param membership_type: WorkspaceMembership value, e.g. WorkspaceMembership.CREATOR
        :return: WorkspaceMember instance
        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/members/add/'),
                                  json=WorkspaceMember(username=username, membership_type=membership_type).to_json(),
                                  expected_status=HTTPStatus.OK)

        return WorkspaceMember.from_json(response.json())

    def change_membership(self, username, membership_type):
        """\
        Changes the membership level for a user.

        :param username: username
        :param membership_type: new membership type, e.g. WorkspaceMembership.CREATOR
        :return: WorkspaceMember instance

        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/members/change/'),
                                  json=WorkspaceMember(username=username, membership_type=membership_type).to_json(),
                                  expected_status=HTTPStatus.OK)

        return WorkspaceMember.from_json(response.json())

    def remove_member(self, username):
        """\
        Removes a member from this workspace.

        :param username: username
        :return: WorkspaceMember instance

        """
        response = self._gf._post(urljoin(self.endpoint, f'{self.api_id}/members/remove/'),
                                  json=WorkspaceMember(username=username, membership_type=None).to_json(),
                                  expected_status=HTTPStatus.OK)

        return WorkspaceMember.from_json(response.json())

    def get_recents(self, limit=100):
        """\
        Gets the most recently created or modified analyses & figures.

        :param limit: maximum number of elements to retrieve.
        :return: Instance of the Recents object
        """
        response = self._gf._get(urljoin(self.endpoint, f'{self.api_id}/recent/?limit={limit}'),
                                         expected_status=HTTPStatus.OK)
        data = response.json()
        analyses = [self._gf.Analysis(**datum) for datum in data.get("analyses", [])]
        figures = [self._gf.Figure(**datum) for datum in data.get("figures", [])]
        return Recents(analyses, figures)


class gf_Analysis(ShareableModelMixin, LogsMixin, ThumbnailMixin):
    """Represents an analysis"""
    # pylint: disable=protected-access

    fields = ["api_id",
              "_shallow",
              "name",
              "description",
              "size_bytes",
              LinkedEntityField("workspace", lambda gf: gf.Workspace, many=False),
              LinkedEntityField("figures", lambda gf: gf.Figure, many=True, derived=True,
                                backlink_property='analysis')] + TIMESTAMP_FIELDS + CHILD_TIMESTAMP_FIELDS
    endpoint = "analysis/"

    def get_figure(self, name, create=True, **kwargs):
        """\
        Finds a figure by name, optionally creating it.

        :param name: name of the figure to find
        :param create: True to create the figure if it doesn't exist.
        :param kwargs: parameters to Figure (e.g. description) if it needs to be created
        :return: Figure instance

        """
        return self.figures.find_or_create(name=name,
                                           default_obj=self._gf.Figure(name=name, **kwargs) if create else None)


class gf_Figure(ShareableModelMixin, ThumbnailMixin):
    """Represents a figure"""
    fields = ["api_id",
              "_shallow",
              "name",
              "description",
              "size_bytes",
              LinkedEntityField("analysis", lambda gf: gf.Analysis, many=False),
              LinkedEntityField("revisions", lambda gf: gf.Revision, many=True,
                                derived=True, backlink_property='figure')
              ] + TIMESTAMP_FIELDS + CHILD_TIMESTAMP_FIELDS
    endpoint = "figure/"


class DataType(abc.ABC):
    """\
    Enum for different types of data we can store inside a FigureRevision.
    """
    DATA_FRAME = "dataframe"
    CODE = "code"
    IMAGE = "image"
    TEXT = "text"


class MetadataProxyField:
    """Field which is embedded inside the JSON metadata"""
    def __init__(self, name, default=None):
        """

        :param name: name of the field
        :param default: default value
        """
        self.name = name
        self.default = default

    def __get__(self, instance, owner):
        """\
        Retrieves the field value.

        :param instance:
        :param owner:
        :return:
        """
        return instance.metadata.get(self.name, self.default)

    def __set__(self, instance, value):
        """\
        Sets the field value.

        :param instance:
        :param value:
        :return:
        """
        if not hasattr(instance, 'metadata') or instance.metadata is None:
            instance.metadata = {}

        instance.metadata[self.name] = value

    def __delete__(self, instance):
        """\
        Deletes the field.

        :param instance:
        :return:
        """
        if instance.metadata is None:
            return
        elif self.name in instance.metadata:
            del instance.metadata[self.name]


class gf_Data(ModelMixin):
    """Represents binary data (e.g. image data, serialized dataframes, etc.)"""
    # pylint: disable=no-member

    SPECIALIZED_TYPES = None
    DATA_TYPE = None

    fields = ["api_id",
              "name",
              "type",
              JSONField("metadata"),
              Base64Field("data")]

    metadata_fields = []

    endpoint = "data/"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = kwargs.pop('type', self.DATA_TYPE)
        self.metadata = kwargs.pop('metadata', {})

        # Assign metadata fields
        for meta in self.metadata_fields:
            if meta.name in kwargs:
                self.metadata[meta.name] = kwargs[meta.name]
            elif meta.name not in self.metadata:
                self.metadata[meta.name] = meta.default

        # Assign a new local ID (unless one already exists)
        if 'local_id' in kwargs:
            self.local_id = kwargs['local_id']

        if self.local_id is None:
            self.local_id = str(uuid4())

    @property
    def local_id(self):
        """Reserved for internal API use"""
        if self.metadata is None or 'local_id' not in self.metadata:
            return None

        return self.metadata['local_id']

    @local_id.setter
    def local_id(self, value):
        """Reserved for internal API use"""
        if self.metadata is None:
            self.metadata = {}

        self.metadata['local_id'] = value

    @property
    def is_shallow(self):
        """True if this is a shallow data object without the byte payload"""
        return self.api_id is not None and self.data is None

    def specialize(self):
        """Creates a specialized data instance, e.g. ImageData, based on the data type"""
        supported_types = {
            DataType.IMAGE: self.client.ImageData,
            DataType.CODE: self.client.CodeData,
            DataType.TEXT: self.client.TextData,
            DataType.DATA_FRAME: self.client.TableData
        }
        if self.type not in supported_types:
            raise ValueError(f"Unsupported data type: {self.type}. No specialized class exists.")

        return supported_types[self.type](api_id=self.api_id,
                                          name=self.name,
                                          type=self.type,
                                          metadata=self.metadata,
                                          data=self.data)

    def __eq__(self, other):
        repr1 = self.to_json()
        repr2 = other.to_json()

        if not set(repr1.keys()) == set(repr2.keys()):
            return False
        else:
            return all(repr1[k] == repr2[k] for k in repr1.keys()
                       if not isinstance(self.fields[k], Timestamp) and k != "api_id")

    def __repr__(self):
        dat = self.to_json()
        dat['data'] = f"<{len(self.data)} bytes>" if self.data else None
        return str(dat)


class gf_ImageData(gf_Data):
    """Binary image data"""
    # pylint: disable=no-member

    DATA_TYPE = DataType.IMAGE

    is_watermarked = MetadataProxyField("is_watermarked", default=False)
    format = MetadataProxyField("format")
    metadata_fields = [is_watermarked, format]

    @property
    def is_interactive(self):
        """True if this figure is interactive, false otherwise"""
        return self.format == "html"

    @property
    def image(self):
        """\
        Returns this image as a PIL.Image object.

        :return: PIL.Image
        """
        if self.is_interactive:
            raise RuntimeError("This is an interactive figure.")

        if self.data is None:
            return None

        bio = io.BytesIO(self.data)
        img = PIL.Image.open(bio)
        img.load()
        return img


class CodeLanguage(abc.ABC):
    """\
    For code data objects, the programming language of the embedded code.
    """
    PYTHON = "Python"
    R = "R"


class gf_CodeData(gf_Data):
    """Serialized code data (it's text, but stored and transmitted as bytes)"""
    # pylint: disable=no-member

    DATA_TYPE = DataType.CODE

    language = MetadataProxyField("language")
    format = MetadataProxyField("format")
    encoding = MetadataProxyField("encoding", default="utf-8")
    metadata_fields = [language, format, encoding]

    def __init__(self, contents=None, **kwargs):
        """\

        :param contents: code string. Will be encoded and stored internally as bytes.
        :param kwargs: same parameters as the Data constructor
        """
        super().__init__(**kwargs)
        if contents is not None:
            self.contents = contents

        self.format = kwargs.get("format", "text")

    @property
    def contents(self):
        """\
        Code contents.

        :return:
        """
        return self.data.decode(self.encoding) if self.data is not None else None

    @contents.setter
    def contents(self, value):
        """\
        Sets the code contents.

        :param value:
        :return:
        """
        self.data = value.encode(self.encoding) if value is not None else None


class gf_TextData(gf_Data):
    """Serialized text data (even though it's text, we store and transmit bytes)"""
    # pylint: disable=no-member

    DATA_TYPE = DataType.TEXT
    encoding = MetadataProxyField("encoding", default="utf-8")
    metadata_fields = [encoding]

    def __init__(self, contents=None, **kwargs):
        super().__init__(**kwargs)
        if contents is not None:
            self.contents = contents

    @property
    def contents(self):
        """\
        Decoded text (a string)
        """
        return self.data.decode(self.encoding) if self.data is not None else None

    @contents.setter
    def contents(self, value):
        """\
        Setter for the text.

        :param value:
        :return:
        """
        self.data = value.encode(self.encoding) if value is not None else None


class gf_TableData(gf_Data):
    """Serialized data frame"""
    # pylint: disable=no-member

    DATA_TYPE = DataType.DATA_FRAME
    format = MetadataProxyField("format")
    encoding = MetadataProxyField("encoding", default="utf-8")
    metadata_fields = [format, encoding]

    def __init__(self, dataframe=None, **kwargs):
        """

        :param dataframe: pd.DataFrame to store. Will be converted to CSV and stored as bytes.
        :param kwargs: same as Data
        """
        super().__init__(**kwargs)
        self.format = "pandas/csv"

        if dataframe is not None:
            self.dataframe = dataframe

    @property
    def dataframe(self):
        """
        Parses the dataframe from the embedded stream of bytes.

        :return:
        """
        return pd.read_csv(io.BytesIO(self.data), encoding=self.encoding) if self.data is not None else None

    @dataframe.setter
    def dataframe(self, value):
        """\
        Stores the dataframe by converting to CSV and saving as bytes.

        :param value: pd.DataFrame instance
        :return:
        """
        self.data = value.to_csv().encode(self.encoding) if value is not None else None


class gf_Revision(ShareableModelMixin, ThumbnailMixin):
    """Represents a figure revision"""
    fields = ["api_id", "revision_index", "size_bytes",
              JSONField("metadata"),
              LinkedEntityField("figure", lambda gf: gf.Figure, many=False),
              DataField("data", lambda gf: gf.Data, many=True),
              ] + TIMESTAMP_FIELDS

    endpoint = "revision/"

    def _replace_data_type(self, data_type, value):
        """\
        Because different data types (image, text, etc.) are stored in a flat list, this is a convenience
        function to only replace data of a certain type and nothing else.

        :param data_type: type of data to replace
        :param value: value to replace it with
        :return: None

        """
        # pylint: disable=access-member-before-definition, attribute-defined-outside-init
        other = [dat for dat in self.data if dat.type != data_type]
        self.data = other + list(value)

    def _restore_data(self, data_copy):
        """\
        Restores data from a copy. This function is necessary because the API only returns shallow data objects
        (no bytes) which, based on the base ModelMixin, will overwrite actual data in this object.

        :param data_copy:
        :return:
        """
        # Fill in API IDs for data objects
        local_id_to_api_id = {obj.local_id: obj.api_id for obj in self.data}
        for datum in data_copy:
            datum.api_id = local_id_to_api_id[datum.local_id]

        self.data = data_copy  # pylint: disable=attribute-defined-outside-init
        return self

    def create(self, update=False):
        """\
        Creates a Revision on the server.

        :param update: True to update the object if it already exists
        :return: self

        """
        self._assert_data_not_shallow()

        # The server will return a shallow copy of the data objects, without the byte payload.
        # Instead of requesting that data separately (slow), we simply preserve the submitted data.
        data_copy = self.data
        super().create(update=update)

        self._restore_data(data_copy)
        return self

    def fetch(self, fetch_data=True):
        """\
        Overrides the default fetch to retrieve the data objects (with byte payload) as well. This override
        is necessary because the API only returns shallow data objects by default.

        :param fetch_data: whether to retrieve full data objects
        :return: self
        """
        super().fetch()
        if fetch_data:
            self.fetch_data()
        return self

    def save(self, create=False, patch=False, silent=False):
        self._assert_data_not_shallow()

        # The server will return a shallow copy of the data objects, without the byte payload.
        # Instead of requesting that data separately (slow), we simply preserve the submitted data.
        data_copy = self.data
        super().save(create=create, patch=patch, silent=silent)

        self._restore_data(data_copy)
        return self

    def _assert_data_not_shallow(self, message="This revision contains shallow data. Did you call fetch_data?"):
        for datum in self.data:
            if datum.is_shallow:
                raise RuntimeError(message)

    def fetch_data(self):
        """Fetches all data objects"""
        if self.data is None:
            return self

        for data in self.data:
            data.fetch()

        return self

    def _update_properties(self, props, parse=True):
        super()._update_properties(props, parse=parse)
        if self.data is None:
            self.data = []

        self.data = [dat.specialize() for dat in self.data]
        return self

    @property
    def revision_url(self):
        """Returns the GoFigr URL for this revision"""
        return f"{self._gf.app_url}/r/{self.api_id}"

    @property
    def image_data(self):
        """Returns only image data (if any)"""
        return [dat for dat in self.data if dat.type == DataType.IMAGE]

    @image_data.setter
    def image_data(self, value):
        self._replace_data_type(DataType.IMAGE, value)

    @property
    def table_data(self):
        """Returns only DataFrame data (if any)"""
        return [dat for dat in self.data if dat.type == DataType.DATA_FRAME]

    @table_data.setter
    def table_data(self, value):
        self._replace_data_type(DataType.DATA_FRAME, value)

    @property
    def code_data(self):
        """Returns only code data (if any)"""
        return [dat for dat in self.data if dat.type == DataType.CODE]

    @code_data.setter
    def code_data(self, value):
        self._replace_data_type(DataType.CODE, value)

    @property
    def text_data(self):
        """Returns only text data (if any)"""
        return [dat for dat in self.data if dat.type == DataType.TEXT]

    @text_data.setter
    def text_data(self, value):
        self._replace_data_type(DataType.TEXT, value)


class gf_ApiKey(ModelMixin):
    """\
    Represents an API key. The field 'token' is the actual token used to authenticate, and is always null
    except at key creation.

    """

    fields = ["api_id",
              "name",
              "token",
              Timestamp("created"),
              Timestamp("last_used"),
              Timestamp("expiry"),
              LinkedEntityField("workspace", lambda gf: gf.Workspace, many=False),]
    endpoint = "api_key/"

    def fetch_and_preserve_token(self):
        """\
        Like fetch(), but preserves the token if present. This is useful because the token is only available
        at creation, so calling fetch() will always set it to None.

        :return:
        """
        # pylint: disable=access-member-before-definition,attribute-defined-outside-init
        tok = self.token
        self.fetch()
        self.token = tok
        return self


class gf_WorkspaceInvitation(ModelMixin):
    """\
    Represents an invitation to join a workspace.

    """
    # pylint: disable=protected-access

    def _require_token(self):
        if self.token is None:
            raise ValueError("This action requires a valid token")

    def fetch(self):
        """\
        Updates all fields from the server. Note that any unsaved local changes will be overwritten.

        :return: self
        """
        if self.token is not None:
            obj = self._gf._get(urljoin(self.endpoint, self.token)).json()
        else:
            self._check_api_id()
            obj = self._gf._get(urljoin(self.endpoint, self.api_id)).json()

        return self._update_properties(obj)

    def delete(self, **kwargs):
        return self.client._delete(urljoin(self.endpoint, self.api_id))

    def accept(self):
        """Accepts this invitation"""
        self._require_token()
        return self.client._post(urljoin(self.endpoint, self.token + "/accept"), json={})

    fields = ["api_id",
              "email",
              "initiator",
              "token",
              "status",
              Timestamp("created"),
              Timestamp("expiry"),
              LinkedEntityField("workspace", lambda gf: gf.Workspace, many=False),
              "membership_type"]
    endpoint = "invitations/workspace/"


class gf_MetadataProxy(ModelMixin):
    """\
    Represents a proxy object for sharing metadata between the GoFigr extension (server-side) and the Jupyter client.

    """
    # pylint: disable=protected-access
    def fetch(self):
        """\
        Updates all fields from the server. Note that any unsaved local changes will be overwritten.

        :return: self
        """
        obj = self._gf._get(urljoin(self.endpoint, self.token)).json()
        return self._update_properties(obj)

    def delete(self, **kwargs):
        return self.client._delete(urljoin(self.endpoint, self.token))

    def save(self, *args, **kwargs):  # pylint: disable=unused-argument
        response = self.client._post(urljoin(self.endpoint, self.token), json={'metadata': self.metadata})
        if response.json() is not None:
            self._update_properties(response.json())
        return self

    def update_metadata(self, metadata):
        """Accepts this invitation"""
        self._require_token()
        return self.client._post(urljoin(self.endpoint, self.token), json=metadata)

    fields = ["api_id",
              "initiator",
              "token",
              Timestamp("created"),
              Timestamp("expiry"),
              Timestamp("updated"),
              JSONField("metadata")]
    endpoint = "metadata/"



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
from collections import Counter
from unittest import TestCase

from gofigr import WorkspaceType
from tests.test_client import make_gf, GfTestCase


def get_action_counts(logs):
    return tuple(sorted(Counter([log.action for log in logs]).items()))


class TestFigures(GfTestCase):
    def test_logs(self):
        gf = make_gf()

        workspace = gf.Workspace(name="Test workspace").create()
        ana = gf.Analysis(name="Test analysis", workspace=workspace).create()
        fig = gf.Figure(name="Test figure", analysis=ana).create()
        rev = gf.Revision(figure=fig, metadata={'test': 1}).create()

        # Update workspace name
        workspace.name = "Updated name"
        workspace.save()

        # Update figure name
        fig.name = "Updated name"
        fig.save()

        # Update revision metadata
        rev.metadata = {'test': 2}
        rev.save()

        # Delete revision
        rev.delete(delete=True)

        # Grab the workspace logs
        logs = [log.fetch() for log in workspace.get_logs()]
        self.assertEqual(len(logs), 8)  # 4 create + 3 update + 1 delete
        self.assertEqual(get_action_counts(logs), (("create", 4), ("delete", 1), ("update", 3)))

        # Grab the analysis logs. Should be the same except the workspace-level actions.
        ana_logs = [log.fetch() for log in ana.get_logs()]
        self.assertEqual(len(ana_logs), 6)  # 4 create + 3 update + 1 delete
        self.assertEqual(get_action_counts(ana_logs), (("create", 3), ("delete", 1), ("update", 2)))



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

The metadata proxy uses the GoFigr server to securely pass information from JavaScript to the GoFigr plugin
running inside the Python kernel. This mechanism enables the frontend to securely communicate regardless
of how Jupyter is running, i.e. it works even with password/token auth and SSH tunneling.

"""
# pylint: disable=global-statement

import multiprocessing
import sys
import time

from queue import Empty as QueueEmpty

from threading import Thread
from urllib.parse import urljoin

from IPython.core.display import Javascript

from gofigr.utils import read_resource_text, read_resource_b64

_CALLBACK_THREAD = None
_STOP_CALLBACK_THREAD = False


QUEUE_MSG_STARTED = "started"
QUEUE_MSG_DONE = "done"


def callback_thread(callback, queue, proxy):
    """Periodically polls the queue for messages, and calls the callback function (if specified)"""
    queue.put(QUEUE_MSG_STARTED)

    while not _STOP_CALLBACK_THREAD:
        proxy.fetch()
        if proxy.updated is not None:  # we have metadata
            callback(proxy)
            queue.put(QUEUE_MSG_DONE)
            return

        time.sleep(0.25)


def wait_for_metadata(queue, timeout):
    """Creates a callable which waits for metadata to be fetched, up to a given timeout (seconds)"""
    def _waiter():
        try:
            res = queue.get(block=True, timeout=timeout)
            if res != QUEUE_MSG_DONE:
                print(f"Unexpected proxy message: {res}", file=sys.stderr)

            queue.task_done()
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"Unable to retrieve metadata from the GoFigr proxy: {e}", file=sys.stderr)

    return _waiter


def run_proxy_async(gf, callback):
    """\
    Starts the GoFigr WebSocket listener in the background. Blocks until the listener is setup and ready
    to process messages.

    :param gf: GoFigr client instance
    :param callback: function to call with received messages
    :return: port that the listener was started on

    """
    global _CALLBACK_THREAD, _STOP_CALLBACK_THREAD

    if _CALLBACK_THREAD is not None:
        _STOP_CALLBACK_THREAD = True
        _CALLBACK_THREAD.join(2.0)

    mp_queue = multiprocessing.JoinableQueue()

    proxy = gf.MetadataProxy().create()

    try:
        _STOP_CALLBACK_THREAD = False
        _CALLBACK_THREAD = Thread(target=callback_thread,
                                  args=(callback, mp_queue, proxy))
        _CALLBACK_THREAD.start()

        res = mp_queue.get(block=True, timeout=5)
        if res != QUEUE_MSG_STARTED:
            raise QueueEmpty()

        mp_queue.task_done()

    except QueueEmpty:
        print("GoFigr JavaScript proxy did not start and functionality may be limited.",
              file=sys.stderr)

    return proxy, wait_for_metadata(mp_queue, timeout=5)


def get_javascript_loader(gf, proxy):
    """Creates a Javascript loader for GoFigr. Intended to be display()'d in the call to configure()"""
    endpoint = urljoin(gf.api_url, "metadata/" + proxy.token)
    loader_body = read_resource_text("gofigr.resources", "loader.js")

    css_b64 = read_resource_b64("gofigr.resources", "gofigr.css")

    loader = f"const endpoint=\"{endpoint}\"; const gofigr_css=\"{css_b64}\"\n" + loader_body
    return Javascript(loader)



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""


# pylint: disable=protected-access
class RevisionContext:
    """\
    Stores publishing context for a revision. We created this class so that we don't have to pass
    a lot of disjointed state around.

    """
    def __init__(self, backend=None, extension=None):
        """\

        :param backend: GoFigr backend which originated this figure
        :param extension: GoFigr Jupyter extension
        """
        self.backend = backend
        self.extension = extension

    @staticmethod
    def get(obj):
        """Gets the revision context associated with a revision. None if not available."""
        if hasattr(obj, '_revision_context'):
            return obj._revision_context
        else:
            return None

    def attach(self, obj):
        """Attaches context to a revision"""
        obj._revision_context = self



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import inspect
from abc import ABC


def get_all_function_arguments(frame):
    """Iterates over all function arguments, including *args and **kwargs"""
    arg_values = inspect.getargvalues(frame[0])

    # Positional arguments
    for arg_name in arg_values.args:
        arg_value = arg_values.locals[arg_name]
        yield arg_value

    # Varargs
    if arg_values.varargs:
        # pylint: disable=use-yield-from
        for arg_value in arg_values.locals[arg_values.varargs]:
            yield arg_value

    # Kwargs
    if arg_values.keywords:
        # pylint: disable=use-yield-from
        for arg_value in arg_values.locals[arg_values.keywords].values():
            yield arg_value


class GoFigrBackend(ABC):
    """Base class for figure backends, e.g. matplotlib or plotly"""
    def is_compatible(self, fig):
        """Returns True if this backend is compatible with a figure"""
        raise NotImplementedError

    def is_interactive(self, fig):
        """Returns True if the figure supports interactive (HTML) output"""
        raise NotImplementedError

    def is_static(self, fig):
        """Returns True if the figure is static (e.g. an image)"""
        return not self.is_interactive(fig)

    def find_figures(self, shell, data):
        """\
        Finds all figures compatible with this backend in the current environment.

        :param shell: IPython shell
        :return: iterator over available figures
        """
        raise NotImplementedError

    def get_default_figure(self, silent=False):
        """"\
        Returns the default/current figure, e.g. plt.gcf() for matplotlib

        :param silent: if True, will suppress warnings/errors if default figure is unavailable
        """
        raise NotImplementedError

    def get_title(self, fig):
        """\
        Extracts figure title

        :param fig: figure to extract title from
        :return: figure title, or None if not available
        """
        raise NotImplementedError

    def figure_to_bytes(self, fig, fmt, params):
        """\
        Given a figure and an image format, converts the figure to binary representation in that format.

        :param fig: figure object
        :param fmt: format, e.g. png
        :param params: backend-specific parameters
        :return: bytes

        """
        raise NotImplementedError

    def figure_to_html(self, fig):
        """\
        Converts a figure to interactive HTML (if supported by the backend)

        :param fig: figure to convert to HTML
        :return: figure as HTML string

        """
        raise NotImplementedError

    def add_interactive_watermark(self, fig, rev, watermark):
        """\
        Adds watermark to a figure using the backend's native objects (if supported by the backend)

        :param fig: figure to watermark
        :param rev: FigureRevision object
        :param watermark: DefaultWatermark instance
        :return: modified figure

        """
        raise NotImplementedError

    def close(self, fig):
        """\
        Closes a figure and prevents it from being displayed.
        :param fig: figure to close
        :return: None
        """
        raise NotImplementedError

    def get_backend_name(self):
        """\
        Gets a human-readable name of this backend.
        :return: backend name, a string
        """
        raise NotImplementedError

    def get_supported_image_formats(self):
        """Gets a list of supported image formats"""
        raise NotImplementedError


def get_backend(figure, backends):
    """\
    Gets the backend compatible with a figure

    :param figure: figure to find a backend for
    :param backends: list of backends to search
    :return: GoFigr backend compatible with the figure, or ValueError if none are compatible.

    """
    for back in backends:
        if back.is_compatible(figure):
            return back

    raise ValueError(f"Could not find a backend compatible with {figure}. "
                     f"Checked: {', '.join([str(back) for back in backends])}")



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import time
from datetime import datetime, timedelta

from dateutil.tz import tz

from gofigr import UnauthorizedError, GoFigr

from tests.test_client import MultiUserTestCase


def has_invite(workspace, invite):
    for inv in workspace.get_invitations():
        if inv.api_id == invite.api_id:
            return True
    return False


class TestMetadataProxy(MultiUserTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n_revisions = 0

    def test_proxy(self):
        gf = self.gf1

        # First make sure we can't create proxies which don't expire or expire too far into the future
        meta = gf.MetadataProxy(expiry=None).create()
        self.assertIsNotNone(meta.expiry)
        self.assertLess((meta.expiry - meta.created).total_seconds(), 24 * 3600)

        meta = gf.MetadataProxy(expiry=datetime.now() + timedelta(days=7)).create()
        self.assertIsNotNone(meta.expiry)
        self.assertLess((meta.expiry - meta.created).total_seconds(), 24 * 3600)

        meta = gf.MetadataProxy(expiry=datetime.now(tz=tz.tzlocal()) + timedelta(seconds=4)).create()

        meta2 = gf.MetadataProxy(token=meta.token).fetch()
        for prop in ['token', 'api_id', 'initiator']:
            self.assertIsNotNone(getattr(meta, prop))
            self.assertIsNotNone(getattr(meta2, prop))
            self.assertEqual(getattr(meta, prop), getattr(meta2, prop))

        self.assertIsNone(meta.updated)
        self.assertIsNone(meta2.updated)

        meta.metadata = {"abc": "xyz"}
        meta.save()

        meta2.fetch()
        self.assertEqual(meta.metadata, {"abc": "xyz"})
        self.assertEqual(meta.metadata, meta2.metadata)

        self.assertIsNotNone(meta.updated)
        self.assertIsNotNone(meta2.updated)
        self.assertEqual(meta.updated, meta2.updated)
        self.assertLess((datetime.now(tz=tz.tzlocal()) - meta.updated).total_seconds(), 10)

        # User 2 should not be able to fetch any of the metadata, even with a valid token
        u2_meta = self.gf2.MetadataProxy(token=meta.token)
        self.assertRaises(UnauthorizedError, lambda: u2_meta.fetch())

        # However, they should be able to update the metadata
        u2_meta.metadata = "123"
        u2_meta.save()

        meta.fetch()
        self.assertEqual(meta.metadata, "123")

        # Likewise for anonymous access
        gfanon = GoFigr(anonymous=True, url=self.gf1.service_url)
        anon_meta = gfanon.MetadataProxy(token=meta.token)
        self.assertRaises(UnauthorizedError, lambda: anon_meta.fetch())

        anon_meta.metadata = "456"
        anon_meta.save()

        meta.fetch()
        self.assertEqual(meta.metadata, "456")

        # Wait for the token to expire
        time_wait = (meta.expiry - meta.created).total_seconds()
        print(f"Waiting {time_wait:.1f}s for token to expire")
        time.sleep(time_wait)

        self.assertRaises(RuntimeError, lambda: meta.fetch())
        self.assertRaises(RuntimeError, lambda: meta.save())



"""\
Copyright (c) 2024, Flagstaff Solutions, LLC
All rights reserved.

"""
from abc import ABC
# pylint: disable=trailing-whitespace

from base64 import b64encode
from uuid import uuid4

import humanize
from IPython.core.display import HTML

try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display

from gofigr.utils import read_resource_b64, read_resource_text


BUTTON_STYLE = "padding-top: 0.25em; padding-bottom: 0.25em; padding-left: 0.5em; padding-right: 0.5em; " + \
               "color: #fff; display: inline-block; font-weight: 400; line-height: 1; " + \
               "text-align: center; vertical-align: middle; cursor: pointer; border: 1px solid transparent; " + \
               "font-size: 0.875rem; border-radius: 0.35rem; margin-top: auto; margin-bottom: auto; "

VIEW_BUTTON_STYLE = BUTTON_STYLE + "background-color: #ed6353; border-color: #ed635e; margin-left: 0.5rem; "
DOWNLOAD_BUTTON_STYLE = BUTTON_STYLE + "background-color: #0d53b1; border-color: #0d53b1; margin-left: 0.5rem; "

COPY_BUTTON_STYLE = BUTTON_STYLE + "background-color: #2c7df5; border-color: #2c7df5; margin-left: 0.5rem; "

WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; border: 1px solid transparent; border-color: #cedcef; padding: 0.75em; " + \
               "border-radius: 0.35rem; flex-wrap: wrap; "

COMPACT_WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; border: 1px solid transparent; border-color: #cedcef; padding: 0.25em; " + \
               "padding-top: 0.5em; " + \
               "border-radius: 0.35rem; flex-wrap: wrap; "

MINIMAL_WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; padding: 0.25em; " + \
               "padding-top: 0.5em; flex-wrap: wrap; "

MESSAGE_STYLE = "padding-top: 0.25rem; "

ROW_STYLE = "width: 100%; display: flex; flex-wrap: wrap;"

BREAK_STYLE = "flex-basis: 100%; height: 0;"

FA_VIEW = """<svg xmlns="http://www.w3.org/2000/svg" height="0.85rem" width="0.85rem" viewBox="0 0 512 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M352 0c-12.9 0-24.6 7.8-29.6 19.8s-2.2 25.7 6.9 34.9L370.7 96 201.4 
265.4c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0L416 141.3l41.4 41.4c9.2 9.2 22.9 11.9 34.9 6.9s19.8-16.6 
19.8-29.6V32c0-17.7-14.3-32-32-32H352zM80 32C35.8 32 0 67.8 0 112V432c0 44.2 35.8 80 80 80H400c44.2 0 80-35.8 
80-80V320c0-17.7-14.3-32-32-32s-32 14.3-32 32V432c0 8.8-7.2 16-16 16H80c-8.8 0-16-7.2-16-16V112c0-8.8 7.2-16 
16-16H192c17.7 0 32-14.3 32-32s-14.3-32-32-32H80z"/></svg>"""

FA_COPY = """<svg xmlns="http://www.w3.org/2000/svg" height="1rem" width="0.85rem" viewBox="0 0 448 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M208 0H332.1c12.7 0 24.9 5.1 33.9 14.1l67.9 67.9c9 9 14.1 21.2 14.1 
33.9V336c0 26.5-21.5 48-48 48H208c-26.5 0-48-21.5-48-48V48c0-26.5 21.5-48 48-48zM48 128h80v64H64V448H256V416h64v48c0 
26.5-21.5 48-48 48H48c-26.5 0-48-21.5-48-48V176c0-26.5 21.5-48 48-48z"/></svg>"""

FA_COPY_LIGHT = FA_COPY.replace("#ffffff", "#0d53b1")

FA_DOWNLOAD = """<svg xmlns="http://www.w3.org/2000/svg" height="1rem" width="1rem" viewBox="0 0 512 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M288 32c0-17.7-14.3-32-32-32s-32 14.3-32 
32V274.7l-73.4-73.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3l128 128c12.5 12.5 32.8 12.5 45.3 
0l128-128c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L288 274.7V32zM64 352c-35.3 0-64 28.7-64 64v32c0 35.3 28.7 64 
64 64H448c35.3 0 64-28.7 64-64V416c0-35.3-28.7-64-64-64H346.5l-45.3 45.3c-25 25-65.5 25-90.5 0L165.5 352H64zm368 
56a24 24 0 1 1 0 48 24 24 0 1 1 0-48z"/></svg>"""


def timestamp_to_local_tz(dt_tz):
    """Converts a datetime to the local timezone"""
    return dt_tz.astimezone(tz=None)


class WidgetBase(ABC):
    """Generates HTML/Javascript for the GoFigr Jupyter widget shown under each figure"""

    def __init__(self, revision):
        """

        :param revision: gf.Revision instance
        """
        self.revision = revision
        self._logo_b64 = None
        self.id = f"gf-widget-{str(uuid4())}"
        self.alert_id = f"{self.id}-alert"

    def get_logo_b64(self):
        """Reads the GoFigr logo and returns it as a base64 string"""
        if self._logo_b64 is not None:
            return self._logo_b64

        self._logo_b64 = read_resource_b64("gofigr.resources", "logo_small.png")
        return self._logo_b64

    def get_download_link(self, label, watermark, image_format="png"):
        """Generates HTML for the download link"""
        matches = [img for img in self.revision.image_data
                   if img.format and img.format.lower() == image_format.lower() and img.is_watermarked == watermark]
        if len(matches) == 0:
            return ""
        else:
            img = matches[0]
            return f"""<a download="{self.revision.figure.name}.{image_format}" 
                          style="{DOWNLOAD_BUTTON_STYLE}"
                          href="data:image/{image_format};base64,{b64encode(img.data).decode('ascii')}">
                          {label}</a>"""

    def _render_template(self, name, mapping):
        """Renders a JavaScript template with the given variable mappings"""
        template = read_resource_text("gofigr.resources", name).replace("\n", "")
        for key, value in mapping.items():
            template = template.replace(key, value)
        template = template.replace("\"", "'")
        return template

    def get_copy_link(self, label, watermark, image_format="png"):
        """Generates HTML/JS for the copy image button"""
        matches = [img for img in self.revision.image_data
                   if img.format and img.format.lower() == image_format.lower() and img.is_watermarked == watermark]
        if len(matches) == 0:
            return ""

        img = matches[0]
        onclick = self._render_template("copy_image.js",
                                        {"_ALERT_ID_": self.alert_id,
                                         "_ERROR_MESSAGE_": "Unable to copy image. "
                                                            "Are you running over HTTPS and/or localhost?",
                                         "_SUCCESS_MESSAGE_": "Image copied to clipboard!",
                                         "_IMAGE_B64_": b64encode(img.data).decode('ascii')})
        return f"""<button onclick="{onclick}" style="{COPY_BUTTON_STYLE}">{label}</button>"""

    def get_text_copy_link(self, label, text):
        """Generates HTML/JS for the copy link button"""
        onclick = read_resource_text("gofigr.resources", "copy_text.js").replace("\n", "")
        onclick = onclick.replace("_TEXT_", text)
        onclick = onclick.replace("_ALERT_ID_", self.alert_id)
        onclick = onclick.replace("_SUCCESS_MESSAGE_", "Link copied to clipboard!")
        onclick = onclick.replace("_ERROR_MESSAGE_", "Unable to copy link. Make sure your Jupyter instance is "
                                                     "running over HTTPS and/or on localhost.")
        onclick = onclick.replace("\"", "'")

        return f"""<span onclick="{onclick}" style="cursor: pointer; color: #0d53b1; ">{label}</span>"""

    def show(self):
        """Displays the widget in Jupyter"""
        raise NotImplementedError


class DetailedWidget(WidgetBase):
    """Generates HTML/Javascript for the GoFigr Jupyter widget shown under each figure"""
    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        copy_url = self.get_text_copy_link(f"<span style='margin-left: 0.5rem;'>"
                                           f"{self.revision.figure.name}</span>"
                                           f"<span style='margin-left: 0.25rem; margin-right: 0.25rem';>"
                                           f"{FA_COPY_LIGHT}</span>",
                                           self.revision.revision_url)

        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
                style='width: 3rem; height: 3rem'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.5rem"}">
                    <span>Successfully published {copy_url} 
                    on {timestamp_to_local_tz(self.revision.created_on).strftime("%x at %X")}.</span> 
                    <span style="margin-left: 0.25rem;"> 
                    Revision size: {humanize.naturalsize(self.revision.size_bytes)}</span>
                    </div>

                    <div style="{ROW_STYLE}">
                    <!-- Logo -->
                    {logo_html}

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="{VIEW_BUTTON_STYLE}">{FA_VIEW}
                    <span> View on GoFigr</span></a>

                    <!-- Download -->
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download</span>", True, "png")}
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download (no watermark)</span>", False, "png")}

                    <!-- Copy -->
                    {self.get_copy_link(FA_COPY + "<span> Copy</span>", True, "png")}
                    {self.get_copy_link(FA_COPY + "<span> Copy (no watermark)</span>", False, "png")}

                    </div>

                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))


class CompactWidget(WidgetBase):
    """Generates a compact GoFigr widget"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
        style='width: 2rem; height: 2rem'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{COMPACT_WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                    <!-- Logo -->
                    {logo_html}               

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="{VIEW_BUTTON_STYLE}">{FA_VIEW}
                    <span> View on GoFigr</span></a>

                    <!-- Download -->
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download</span>", True, "png")}
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download (no watermark)</span>", False, "png")}

                    <!-- Copy -->
                    {self.get_copy_link(FA_COPY + "<span> Copy</span>", True, "png")}
                    {self.get_copy_link(FA_COPY + "<span> Copy (no watermark)</span>", False, "png")}

                    </div>

                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))


class MinimalWidget(WidgetBase):
    """Generates a compact GoFigr widget"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
        style='width: 2rem; height: 2rem; margin-right: 0.5rem;'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{MINIMAL_WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                    <!-- Logo -->
                    {logo_html}               

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="margin-top: auto; margin-bottom: auto;">View on GoFigr</a>
                    </div>
                    
                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))



"""\
Copyright (c) 2022, Flagstaff Solutions, LLC
All rights reserved.

"""
# pylint: disable=cyclic-import, no-member, global-statement, protected-access, wrong-import-order, ungrouped-imports
# pylint: disable=too-many-locals

import inspect
import io
import json
import os
import pickle
import sys
from collections import namedtuple
from functools import wraps
from uuid import UUID

import PIL
import six

from gofigr import GoFigr, API_URL, UnauthorizedError
from gofigr.annotators import CellIdAnnotator, SystemAnnotator, CellCodeAnnotator, \
    PipFreezeAnnotator, NotebookMetadataAnnotator, EnvironmentAnnotator, BackendAnnotator, HistoryAnnotator
from gofigr.backends import get_backend, GoFigrBackend
from gofigr.backends.matplotlib import MatplotlibBackend
from gofigr.backends.plotly import PlotlyBackend
from gofigr.context import RevisionContext
from gofigr.proxy import run_proxy_async, get_javascript_loader
from gofigr.profile import MeasureExecution
from gofigr.watermarks import DefaultWatermark
from gofigr.widget import DetailedWidget

try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display


PY3DMOL_PRESENT = False
if sys.version_info >= (3, 8):
    try:
        import py3Dmol  # pylint: disable=unused-import
        from gofigr.backends.py3dmol import Py3DmolBackend
        PY3DMOL_PRESENT = True
    except ModuleNotFoundError:
        pass


DISPLAY_TRAP = None


def _mark_as_published(fig):
    """Marks the figure as published so that it won't be re-published again."""
    fig._gf_is_published = True
    return fig


def _is_published(fig):
    """Returns True iff the figure has already been published"""
    return getattr(fig, "_gf_is_published", False)


class GfDisplayPublisher:
    """\
    Custom IPython DisplayPublisher which traps all calls to publish() (e.g. when display(...) is called).

    """
    def __init__(self, pub):
        """

        :param pub: Publisher to wrap around. We delegate all calls to this publisher unless trapped.
        """
        self.pub = pub

    def publish(self, data, *args, **kwargs):
        """
        IPython calls this method whenever it needs data displayed. Our function traps the call
        and calls DISPLAY_TRAP instead, giving it an option to suppress the figure from being displayed.

        We use this trap to publish the figure if auto_publish is True. Suppression is useful
        when we want to show a watermarked version of the figure, and prevents it from being showed twice (once
        with the watermark inside the trap, and once without in the originating call).

        :param data: dictionary of mimetypes -> data
        :param args: implementation-dependent
        :param kwargs: implementation-dependent
        :return: None

        """

        # Python doesn't support assignment to variables in closure scope, so we use a mutable list instead
        is_suppressed = [False]
        def suppress_display():
            is_suppressed[0] = True

        if DISPLAY_TRAP is not None:
            trap = DISPLAY_TRAP
            with SuppressDisplayTrap():
                trap(data, suppress_display=suppress_display)

        if not is_suppressed[0]:
            self.pub.publish(data, *args, **kwargs)

    def __getattr__(self, item):
        """\
        Delegates to self.pub

        :param item:
        :return:
        """
        if item == "pub":
            return super().__getattribute__(item)

        return getattr(self.pub, item)

    def __setattr__(self, key, value):
        """\
        Delegates to self.pub

        :param key:
        :param value:
        :return:
        """
        if key == "pub":
            super().__setattr__(key, value)

        return setattr(self.pub, key, value)

    def clear_output(self, *args, **kwargs):
        """IPython's clear_output. Defers to self.pub"""
        return self.pub.clear_output(*args, **kwargs)


class SuppressDisplayTrap:
    """\
    Context manager which temporarily suspends all display traps.
    """
    def __init__(self):
        self.trap = None

    def __enter__(self):
        global DISPLAY_TRAP
        self.trap = DISPLAY_TRAP
        DISPLAY_TRAP = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        global DISPLAY_TRAP
        DISPLAY_TRAP = self.trap
        self.trap = None


class _GoFigrExtension:
    """\
    Implements the main Jupyter extension functionality. You will not want to instantiate this class directly.
    Instead, please call get_extension().
    """
    def __init__(self, ip,
                 auto_publish=False,
                 notebook_metadata=None):
        """\

        :param ip: iPython shell instance
        :param auto_publish: whether to auto-publish figures
        :param pre_run_hook: function to use as a pre-run hook
        :param post_execute_hook: function to use as a post-execute hook
        :param notebook_metadata: information about the running notebook, as a key-value dictionary

        """
        self.shell = ip
        self.auto_publish = auto_publish
        self.cell = None
        self.notebook_metadata = notebook_metadata

        self.gf = None  # active GF object
        self.workspace = None  # current workspace
        self.analysis = None  # current analysis
        self.publisher = None  # current Publisher instance
        self.wait_for_metadata = None  # callable which waits for metadata to become available

        self.deferred_revisions = []

    def display_trap(self, data, suppress_display):
        """\
         Called whenever *any* code inside the Jupyter session calls display().
        :param data: dictionary of MIME types
        :param suppress_display: callable with no arguments. Call to prevent the originating figure from being shown.
        :return: None

        """
        if self.auto_publish:
            self.publisher.auto_publish_hook(self, data, suppress_display)

    def add_to_deferred(self, rev):
        """\
        Adds a revision to a list of deferred revisions. Such revisions will be annotated in the post_run_cell
        hook, and re-saved.

        This functionality exists because it's possible to load the GoFigr extension and publish figures in the same
        cell, in which case GoFigr will not receive the pre_run_cell hook and will not have access to cell information
        when the figure is published. This functionality allows us to obtain the cell information after it's run
        (in the post_run_cell hook), re-run annotators, and update the figure with full annotations.

        :param rev: revision to defer
        :return: None
        """
        if rev not in self.deferred_revisions:
            self.deferred_revisions.append(rev)

    def check_config(self):
        """Ensures the plugin has been configured for use"""
        props = ["gf", "workspace", "analysis", "publisher"]
        for prop in props:
            if getattr(self, prop, None) is None:
                raise RuntimeError("GoFigr not configured. Please call configure() first.")

    def pre_run_cell(self, info):
        """\
        Default pre-run cell hook.

        :param info: Cell object
        :return:None

        """
        self.cell = info

    def post_run_cell(self, result):
        """Post run cell hook.

        :param result: ExecutionResult
        :return: None

        """
        self.cell = result.info

        if self.notebook_metadata is None and self.wait_for_metadata is not None:
            self.wait_for_metadata()
            self.wait_for_metadata = None

        while len(self.deferred_revisions) > 0:
            rev = self.deferred_revisions.pop(0)
            rev = self.publisher.annotate(rev)
            rev.save(silent=True)

        self.cell = None

    def _register_handler(self, event_name, handler):
        """Inserts a handler at the beginning of the list while avoiding double-insertions"""
        handlers = [handler]
        for hnd in self.shell.events.callbacks[event_name]:
            self.shell.events.unregister(event_name, hnd)
            if hnd != handler:  # in case it's already registered, skip it
                handlers.append(hnd)

        for hnd in handlers:
            self.shell.events.register(event_name, hnd)

    def unregister(self):
        """\
        Unregisters all hooks, effectively disabling the plugin.

        """
        try:
            self.shell.events.unregister('pre_run_cell', self.pre_run_cell)
        except ValueError:
            pass

        try:
            self.shell.events.unregister('post_run_cell', self.post_run_cell)
        except ValueError:
            pass

    def register_hooks(self):
        """\
        Register all hooks with Jupyter.

        :return: None
        """
        global DISPLAY_TRAP
        DISPLAY_TRAP = self.display_trap

        self._register_handler('pre_run_cell', self.pre_run_cell)
        self._register_handler('post_run_cell', self.post_run_cell)

        native_display_publisher = self.shell.display_pub
        if not isinstance(native_display_publisher, GfDisplayPublisher):
            self.shell.display_pub = GfDisplayPublisher(native_display_publisher)


_GF_EXTENSION = None  # GoFigrExtension global


def require_configured(func):
    """\
    Decorator which throws an exception if configure() has not been called yet.

    :param func:
    :return:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if _GF_EXTENSION is None:
            raise RuntimeError("Please load the extension: %load_ext gofigr")
        _GF_EXTENSION.check_config()

        return func(*args, **kwargs)

    return wrapper


@require_configured
def get_extension():
    """Returns the GoFigr Jupyter extension instance"""
    return _GF_EXTENSION


def _load_ipython_extension(ip):
    """\
    Loads the Jupyter extension. Aliased to "load_ipython_extension" (no leading underscore) in the main init.py file.

    :param ip: IPython shell
    :return: None

    """
    global _GF_EXTENSION
    if _GF_EXTENSION is not None:
        _GF_EXTENSION.unregister()

    _GF_EXTENSION = _GoFigrExtension(ip)
    _GF_EXTENSION.register_hooks()


def parse_uuid(val):
    """\
    Attempts to parse a UUID, returning None if input is not a valid UUID.

    :param val: value to parse
    :return: UUID (as a string) or None

    """
    try:
        return str(UUID(val))
    except ValueError:
        return None


ApiId = namedtuple("ApiId", ["api_id"])

class FindByName:
    """\
    Used as argument to configure() to specify that we want to find an analysis/workspace by name instead
    of using an API ID
    """
    def __init__(self, name, description=None, create=False):
        self.name = name
        self.description = description
        self.create = create

    def __repr__(self):
        return f"FindByName(name={self.name}, description={self.description}, create={self.create})"


def parse_model_instance(model_class, value, find_by_name):
    """\
    Parses a model instance from a value, e.g. the API ID or a name.

    :param model_class: class of the model, e.g. gf.Workspace
    :param value: value to parse into a model instance
    :param find_by_name: callable to find the model instance by name
    :return: model instance

    """
    if isinstance(value, model_class):
        return value
    elif isinstance(value, str):
        return model_class(api_id=value)
    elif isinstance(value, ApiId):
        return model_class(api_id=value.api_id)
    elif isinstance(value, FindByName):
        return find_by_name(value)
    else:
        return ValueError(f"Unsupported target specification: {value}. Please specify an API ID, or use FindByName.")


DEFAULT_ANNOTATORS = (NotebookMetadataAnnotator, EnvironmentAnnotator, CellIdAnnotator, CellCodeAnnotator,
                      SystemAnnotator, PipFreezeAnnotator, BackendAnnotator, HistoryAnnotator)
DEFAULT_BACKENDS = (MatplotlibBackend, PlotlyBackend)
if PY3DMOL_PRESENT:
    DEFAULT_BACKENDS = DEFAULT_BACKENDS + (Py3DmolBackend,)


class Publisher:
    """\
    Publishes revisions to the GoFigr server.
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 gf,
                 annotators,
                 backends,
                 watermark=None,
                 image_formats=("png", "eps", "svg"),
                 interactive=True,
                 default_metadata=None,
                 clear=True,
                 save_pickle=True,
                 widget_class=DetailedWidget):
        """

        :param gf: GoFigr instance
        :param annotators: revision annotators
        :param backends: figure backends, e.g. MatplotlibBackend
        :param watermark: watermark generator, e.g. QRWatermark()
        :param image_formats: image formats to save by default
        :param interactive: whether to publish figure HTML if available
        :param clear: whether to close the original figures after publication. If False, Jupyter will display
        both the input figure and the watermarked output. Default behavior is to close figures.
        :param save_pickle: if True, will save the figure in pickle format in addition to any of the image formats
        :param widget_class: Widget type to show, e.g. DetailedWidget or CompactWidget. It will appear below the
        published figure

        """
        self.gf = gf
        self.watermark = watermark or DefaultWatermark()
        self.annotators = annotators
        self.backends = backends
        self.image_formats = image_formats
        self.interactive = interactive
        self.clear = clear
        self.default_metadata = default_metadata
        self.save_pickle = save_pickle
        self.widget_class = widget_class

    def auto_publish_hook(self, extension, data, suppress_display=None):
        """\
        Hook for automatically publishing figures without an explicit call to publish().

        :param extension: GoFigrExtension instance
        :param data: data being published. This will usually be a dictionary of mime formats.
        :param native_publish: callable which will publish the figure using the native backend

        :return: None
        """
        for backend in self.backends:
            compatible_figures = list(backend.find_figures(extension.shell, data))
            for fig in compatible_figures:
                if not _is_published(fig):
                    self.publish(fig=fig, backend=backend, suppress_display=suppress_display)

    @staticmethod
    def _resolve_target(gf, fig, target, backend):
        ext = get_extension()

        if target is None:
            # Try to get the figure's title
            fig_name = backend.get_title(fig)
            if fig_name is None:
                print("Your figure doesn't have a title and will be published as 'Anonymous Figure'. "
                      "To avoid this warning, set a figure title or manually call publish() with a target figure. "
                      "See https://gofigr.io/docs/gofigr-python/latest/start.html#publishing-your-first-figure for "
                      "an example.", file=sys.stderr)
                fig_name = "Anonymous Figure"

            sys.stdout.flush()
            return ext.analysis.get_figure(fig_name, create=True)
        else:
            return parse_model_instance(gf.Figure,
                                        target,
                                        lambda search: ext.analysis.get_figure(name=search.name,
                                                                               description=search.description,
                                                                               create=search.create))

    def _get_pickle_data(self, gf, fig):
        if not self.save_pickle:
            return []

        try:
            bio = io.BytesIO()
            pickle.dump(fig, bio)
            bio.seek(0)

            return [gf.ImageData(name="figure", format="pickle",
                                 data=bio.getvalue(),
                                 is_watermarked=False)]
        except Exception as e: # pylint: disable=broad-exception-caught
            print(f"WARNING: We could not obtain the figure in pickle format: {e}", file=sys.stderr)
            return []

    def _get_image_data(self, gf, backend, fig, rev, image_options):
        """\
        Extracts ImageData in various formats.

        :param gf: GoFigr instance
        :param backend: backend to use
        :param fig: figure object
        :param rev: Revision object
        :param image_options: backend-specific parameters
        :return: tuple of: list of ImageData objects, watermarked image to display

        """
        if image_options is None:
            image_options = {}

        image_to_display = None
        image_data = []
        for fmt in self.image_formats:
            if fmt.lower() not in backend.get_supported_image_formats():
                continue

            if fmt.lower() == "png":
                img = PIL.Image.open(io.BytesIO(backend.figure_to_bytes(fig, fmt, image_options)))
                img.load()
                watermarked_img = self.watermark.apply(img, rev)
            else:
                watermarked_img = None

            # First, save the image without the watermark
            try:
                image_data.append(gf.ImageData(name="figure",
                                               format=fmt,
                                               data=backend.figure_to_bytes(fig, fmt, image_options),
                                               is_watermarked=False))
            except Exception as e:  # pylint: disable=broad-exception-caught
                print(f"WARNING: We could not obtain the figure in {fmt.upper()} format: {e}", file=sys.stderr)
                continue

            # Now, save the watermarked version (if available)
            if watermarked_img is not None:
                bio = io.BytesIO()
                watermarked_img.save(bio, format=fmt)
                img_data = gf.ImageData(name="figure", format=fmt, data=bio.getvalue(),
                                        is_watermarked=True)
                image_data.append(img_data)

                if fmt.lower() == 'png':
                    image_to_display = img_data

        if self.interactive and backend.is_interactive(fig):
            image_data.append(gf.ImageData(name="figure", format="html",
                                           data=backend.figure_to_html(fig).encode('utf-8'),
                                           is_watermarked=False))

            wfig = backend.add_interactive_watermark(fig, rev, self.watermark)
            html_with_watermark = gf.ImageData(name="figure", format="html",
                                               data=backend.figure_to_html(wfig).encode('utf-8'),
                                               is_watermarked=True)
            image_data.append(html_with_watermark)
            image_to_display = wfig  # display the native Figure

        image_data.extend(self._get_pickle_data(gf, fig))

        return image_data, image_to_display

    def annotate(self, rev):
        """
        Annotates a FigureRevision using self.annotators.
        :param rev: revision to annotate
        :return: annotated revision

        """
        for annotator in self.annotators:
            with MeasureExecution(annotator.__class__.__name__):
                annotator.annotate(rev)
        return rev

    def _infer_figure_and_backend(self, fig, backend):
        """\
        Given a figure and a backend where one of the values could be null, returns a complete set
        of a figure to publish and a matching backend.

        :param fig: figure to publish. None to publish the default for the backend
        :param backend: backend to use. If None, will infer from figure
        :return: tuple of figure and backend
        """
        if fig is None and backend is None:
            raise ValueError("You did not specify a figure to publish.")
        elif fig is not None and backend is not None:
            return fig, backend
        elif fig is None and backend is not None:
            fig = backend.get_default_figure()

            if fig is None:
                raise ValueError("You did not specify a figure to publish, and the backend does not have "
                                 "a default.")
        else:
            backend = get_backend(fig, self.backends)

        return fig, backend

    def publish(self, fig=None, target=None, gf=None, dataframes=None, metadata=None,
                backend=None, image_options=None, suppress_display=None):
        """\
        Publishes a revision to the server.

        :param fig: figure to publish. If None, we'll use plt.gcf()
        :param target: Target figure to publish this revision under. Can be a gf.Figure instance, an API ID, \
        or a FindByName instance.
        :param gf: GoFigure instance
        :param dataframes: dictionary of dataframes to associate & publish with the figure
        :param metadata: metadata (JSON) to attach to this revision
        usage this will cause Jupyter to print the whole object which we don't want.
        :param backend: backend to use, e.g. MatplotlibBackend. If None it will be inferred automatically based on \
        figure type
        :param image_options: backend-specific params passed to backend.figure_to_bytes
        :param suppress_display: if used in an auto-publish hook, this will contain a callable which will
        suppress the display of this figure using the native IPython backend.

        :return: FigureRevision instance

        """
        # pylint: disable=too-many-branches
        ext = get_extension()
        gf = gf if gf is not None else ext.gf
        fig, backend = self._infer_figure_and_backend(fig, backend)

        with MeasureExecution("Resolve target"):
            target = self._resolve_target(gf, fig, target, backend)
            if getattr(target, 'revisions', None) is None:
                target.fetch()

        combined_meta = self.default_metadata if self.default_metadata is not None else {}
        if metadata is not None:
            combined_meta.update(metadata)

        context = RevisionContext(backend=backend, extension=ext)
        with MeasureExecution("Bare revision"):
            # Create a bare revision first to get the API ID
            rev = gf.Revision(figure=target, metadata=combined_meta)
            target.revisions.create(rev)

            context.attach(rev)

        deferred = False
        if _GF_EXTENSION.cell is None:
            deferred = True
            get_extension().add_to_deferred(rev)

        with MeasureExecution("Image data"):
            rev.image_data, image_to_display = self._get_image_data(gf, backend, fig, rev, image_options)

        if image_to_display is not None:
            with SuppressDisplayTrap():
                if isinstance(image_to_display, gf.ImageData):
                    display(image_to_display.image)
                else:
                    display(image_to_display)

            if suppress_display is not None:
                suppress_display()

        if dataframes is not None:
            table_data = []
            for name, frame in dataframes.items():
                table_data.append(gf.TableData(name=name, dataframe=frame))

            rev.table_data = table_data

        if not deferred:
            with MeasureExecution("Annotators"):
                # Annotate the revision
                self.annotate(rev)

        with MeasureExecution("Final save"):
            rev.save(silent=True)

            # Calling .save() above will update internal properties based on the response from the server.
            # In our case, this will result in rev.figure becoming a shallow object with just the API ID. Here
            # we restore it from our cached copy, to avoid a separate API call.
            rev.figure = target

        _mark_as_published(fig)

        if self.clear:
            backend.close(fig)

        with SuppressDisplayTrap():
            self.widget_class(rev).show()

        return rev


def from_config_or_env(env_prefix, config_path):
    """\
    Decorator that binds function arguments in order of priority (most important first):
    1. args/kwargs
    2. environment variables
    3. config file
    4. function defaults

    :param env_prefix: prefix for environment variables. Variables are assumed to be named \
    `<prefix> + <name of function argument in all caps>`, e.g. if prefix is ``MYAPP`` and function argument \
    is called host_name, we'll look for an \
    environment variable named ``MYAPP_HOST_NAME``.
    :param config_path: path to the JSON config file. Function arguments will be looked up using their verbatim names.
    :return: decorated function

    """
    def decorator(func):
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            # Read config file, if it exists
            if os.path.exists(config_path):
                with open(config_path, 'r', encoding='utf-8') as f:
                    try:
                        config_file = json.load(f)
                    except Exception as e:
                        raise RuntimeError(f"Error parsing configuration file {config_path}") from e
            else:
                config_file = {}

            sig = inspect.signature(func)
            param_values = sig.bind_partial(*args, **kwargs).arguments
            for param_name in sig.parameters:
                env_name = f'{env_prefix}{param_name.upper()}'
                if param_name in param_values:
                    continue  # value supplied through args/kwargs: ignore env variables and the config file.
                elif env_name in os.environ:
                    param_values[param_name] = os.environ[env_name]
                elif param_name in config_file:
                    param_values[param_name] = config_file[param_name]

            return func(**param_values)

        return wrapper

    return decorator


def find_workspace_by_name(gf, search):
    """\
    Finds a workspace by name.

    :param gf: GoFigr client
    :param search: FindByName instance
    :return: a Workspace object

    """
    matches = [wx for wx in gf.workspaces if wx.name == search.name]
    if len(matches) == 0:
        if search.create:
            wx = gf.Workspace(name=search.name, description=search.description)
            wx.create()
            print(f"Created a new workspace: {wx.api_id}")
            return wx
        else:
            raise RuntimeError(f'Could not find workspace named "{search.name}"')
    elif len(matches) > 1:
        raise RuntimeError(f'Multiple (n={len(matches)}) workspaces match name "{search.name}". '
                           f'Please use an API ID instead.')
    else:
        return matches[0]


def proxy_callback(result):
    """Proxy callback"""
    if result is not None and hasattr(result, 'metadata'):
        _GF_EXTENSION.notebook_metadata = result.metadata


def _make_backend(backend):
    if isinstance(backend, GoFigrBackend):
        return backend
    else:
        return backend()


# pylint: disable=too-many-arguments, too-many-locals
@from_config_or_env("GF_", os.path.join(os.environ['HOME'], '.gofigr'))
def configure(username=None, password=None,
              api_key=None,
              workspace=None, analysis=None, url=API_URL,
              default_metadata=None, auto_publish=True,
              watermark=None, annotators=DEFAULT_ANNOTATORS,
              notebook_name=None, notebook_path=None,
              backends=DEFAULT_BACKENDS,
              widget_class=DetailedWidget,
              save_pickle=True):
    """\
    Configures the Jupyter plugin for use.

    :param username: GoFigr username (if used instead of API key)
    :param password: GoFigr password (if used instead of API key)
    :param api_key: API Key (if used instead of username and password)
    :param url: API URL
    :param workspace: one of: API ID (string), ApiId instance, or FindByName instance
    :param analysis: one of: API ID (string), ApiId instance, or FindByName instance
    :param default_metadata: dictionary of default metadata values to save for each revision
    :param auto_publish: if True, all figures will be published automatically without needing to call publish()
    :param watermark: custom watermark instance (e.g. DefaultWatermark with custom arguments)
    :param annotators: list of annotators to use. Default: DEFAULT_ANNOTATORS
    :param notebook_name: name of the notebook (if you don't want it to be inferred automatically)
    :param notebook_path: path to the notebook (if you don't want it to be inferred automatically)
    :param backends: backends to use (e.g. MatplotlibBackend, PlotlyBackend)
    :param widget_class: Widget type to show, e.g. DetailedWidget or CompactWidget. It will appear below the
        published figure
    :param save_pickle: if True, will save the figure in pickle format in addition to any of the image formats

    :return: None

    """
    extension = _GF_EXTENSION

    if isinstance(auto_publish, str):
        auto_publish = auto_publish.lower() == "true"  # in case it's coming from an environment variable

    with MeasureExecution("Login"):
        gf = GoFigr(username=username, password=password, url=url, api_key=api_key)

    if workspace is None:
        if gf.primary_workspace is not None:
            workspace = gf.primary_workspace
        elif len(gf.workspaces) == 1:  # this will happen if we're using a scoped API token
            workspace = gf.workspaces[0]
            print(f"Defaulting to workspace \"{workspace.name}\" ({workspace.api_id})")
        else:
            raise ValueError("Please specify a workspace")
    else:
        workspace = parse_model_instance(gf.Workspace, workspace, lambda search: find_workspace_by_name(gf, search))

    with MeasureExecution("Fetch workspace"):
        try:
            workspace.fetch()
        except UnauthorizedError as e:
            raise UnauthorizedError(f"Permission denied for workspace {workspace.api_id}. "
                                    f"Are you using a restricted API key?") from e

    if analysis is None:
        raise ValueError("Please specify an analysis")
    else:
        with MeasureExecution("Find analysis"):
            analysis = parse_model_instance(gf.Analysis, analysis,
                                            lambda search: workspace.get_analysis(name=search.name,
                                                                                  description=search.description,
                                                                                  create=search.create))

    with MeasureExecution("Fetch analysis"):
        analysis.fetch()

    if default_metadata is None:
        default_metadata = {}

    if notebook_path is not None:
        default_metadata['notebook_path'] = notebook_path

    if notebook_name is not None:
        default_metadata['notebook_name'] = notebook_name

    publisher = Publisher(gf,
                          default_metadata=default_metadata,
                          watermark=watermark,
                          annotators=[make_annotator(extension) for make_annotator in annotators],
                          backends=[_make_backend(bck) for bck in backends],
                          widget_class=widget_class,
                          save_pickle=save_pickle)
    extension.gf = gf
    extension.analysis = analysis
    extension.workspace = workspace
    extension.publisher = publisher
    extension.auto_publish = auto_publish

    proxy, extension.wait_for_metadata = run_proxy_async(gf, proxy_callback)

    with SuppressDisplayTrap():
        display(get_javascript_loader(gf, proxy))


@require_configured
def publish(fig=None, backend=None, **kwargs):
    """\
    Publishes a figure. See :func:`gofigr.jupyter.Publisher.publish` for a list of arguments. If figure and backend
    are both None, will publish default figures across all available backends.

    :param fig: figure to publish
    :param backend: backend to use
    :param kwargs:
    :return:

    """
    ext = get_extension()

    if fig is None and backend is None:
        # If no figure and no backend supplied, publish default figures across all available backends
        for available_backend in ext.publisher.backends:
            fig = available_backend.get_default_figure(silent=True)
            if fig is not None:
                ext.publisher.publish(fig=fig, backend=available_backend, **kwargs)
    else:
        ext.publisher.publish(fig=fig, backend=backend, **kwargs)


@require_configured
def get_gofigr():
    """Gets the active GoFigr object."""
    return get_extension().gf


@require_configured
def load_pickled_figure(api_id):
    """\
    Unpickles a GoFigr revision and returns it as a backend-specific Python object, e.g. a plt.Figure if
    the figure was generated with matplotlib. Throws a RuntimeException if the figure is not found or does
    not have pickle data.

    :param api_id: API ID of the revision
    :return: backend-dependent figure object, e.g. plt.Figure().

    """
    gf = get_gofigr()
    rev = gf.Revision(api_id=api_id).fetch(fetch_data=False)
    for data in rev.image_data:
        if data.format == "pickle":
            data.fetch()

            fig = pickle.load(io.BytesIO(data.data))
            return _mark_as_published(fig)

    raise RuntimeError("This revision doesn't have pickle data.")
